import os
import mimetypes
from falcon_rest.conf import settings
from falcon_rest.logger import log
from . import ImproperlyConfigured
from . abstract_backend import AbstractBackEnd


FOUND_PATH_IN_CASE_OF_FILE_EXISTS_TRIES_CNT = 100


class FileSystemBackEnd(AbstractBackEnd):
    def __init__(self, get_file_name=None, upload_to=None):
        super(FileSystemBackEnd, self).__init__(get_file_name)
        try:
            self.upload_to = (upload_to or settings.FILES['upload_to'])
            if not self.upload_to:
                raise ImproperlyConfigured
        except (AttributeError, KeyError, ImproperlyConfigured):
            raise ImproperlyConfigured(
                'Neither upload_to passed nor settings.FILES[\'upload_to\'] set or set to empty string'
            )

    def get_path(self):
        path = os.path.join(self.upload_to, self.belongs_to_table, self.belongs_to_field)

        if not os.path.isdir(path):
            try:
                os.makedirs(path)
            except PermissionError:
                log.error('Can\'t create directory {}'.format(path))
                raise PermissionError

        return path

    def write(self, tmp_file):
        file_path = os.path.join(self.get_path(), self.get_file_name(tmp_file))
        with open(self.get_file_path_in_case_already_exists(file_path), 'wb') as fp:
            fp.write(tmp_file.stream.read())

        return file_path

    def read(self, file_obj):
        return (
            self.get_stream(file_obj), self.get_stream_length(file_obj), os.path.basename(file_obj.path),
            self.get_mimetype(file_obj)
        )

    # noinspection PyMethodMayBeStatic
    def get_mimetype(self, file_obj):
        return mimetypes.guess_type(file_obj.path)[0]

    # noinspection PyMethodMayBeStatic
    def get_stream(self, file_obj):
        return open(file_obj.path, 'rb')

    # noinspection PyMethodMayBeStatic
    def get_stream_length(self, file_obj):
        return os.path.getsize(file_obj.path)

    # noinspection PyMethodMayBeStatic
    def get_filename(self, file_obj):
        return os.path.basename(file_obj.path)

    @staticmethod
    def get_file_path_in_case_already_exists(file_path):
        if not os.path.isfile(file_path):
            return file_path

        path, ext = os.path.splitext(file_path)
        for i in range(1, FOUND_PATH_IN_CASE_OF_FILE_EXISTS_TRIES_CNT + 1):
            res = '{}_{}{}'.format(path, i, ext)
            if not os.path.isfile(res):
                return res
