import httplib2
import os
import io
from datetime import datetime
from apiclient import discovery
from oauth2client.service_account import ServiceAccountCredentials
from googleapiclient.http import MediaIoBaseUpload, MediaIoBaseDownload
from falcon_rest.conf import settings


DEFAULT_SCOPES = 'https://www.googleapis.com/auth/drive'
CLIENT_SECRET_FILE = 'client_secret.json'
APPLICATION_NAME = settings.GOOGLE_DRIVE['application_name']
SERVICE_LIFETIME = settings.GOOGLE_DRIVE['service_lifetime']


class GoogleDriveApi(object):
    service = None
    service_created_at = datetime.now()

    def __init__(
            self, client_secret_file, credential_dir, application_name, service_lifetime, log, scopes=DEFAULT_SCOPES
    ):
        self.scopes = scopes
        self.client_secret_file = client_secret_file
        self.credential_dir = credential_dir
        self.application_name = application_name
        self.service_lifetime = float(service_lifetime)
        self.log = log

        if not os.path.exists(self.credential_dir):
            os.makedirs(self.credential_dir)

    def get_service(self):
        current_time = datetime.now()
        if self.service:
            if self.service_created_at:
                if (current_time - self.service_created_at).total_seconds() < self.service_lifetime:
                    return self.service

        # credentials = self.get_credentials()
        credentials = ServiceAccountCredentials.from_json_keyfile_name(self.client_secret_file, self.scopes)
        http = credentials.authorize(httplib2.Http())
        self.log.debug('GoogleDriveApi: building service')
        self.service = discovery.build('drive', 'v3', http=http)
        return self.service

    def upload(self, stream, filename, mimetype, folder_id=None):
        metadata = dict(name=filename)
        if folder_id:
            metadata.update(dict(parents=[folder_id]))

        self.log.debug('GoogleDriveApi: uploading file')
        return self.get_service().files().create(
            body=metadata, media_body=MediaIoBaseUpload(stream, mimetype), fields='id'
        ).execute().get('id')

    def download(self, file_id):
        fh = io.BytesIO()
        self.log.debug('GoogleDriveApi: downloading file')
        downloader = MediaIoBaseDownload(fh, self.get_service().files().get_media(fileId=file_id))
        done = False
        while done is False:
            status, done = downloader.next_chunk()

        fh.seek(0, 0)
        return fh

    def get_file_info(self, file_id, fields=('name', 'size', 'mimeType')):
        kwargs = dict(fileId=file_id)
        if fields:
            kwargs.update(dict(fields=','.join(fields)))

        self.log.debug('GoogleDriveApi: getting file info')
        return self.get_service().files().get(**kwargs).execute()

    def get_folder_id_by_path(self, path, root_folder_id):
        folder_id = root_folder_id

        for folder_name in path:
            folders = self.list_folders(folder_name, folder_id)
            if not folders:
                self.log.debug('GoogleDriveApi: no folder {} found under {}'.format(folder_name, folder_id))
                return None
            folder_id = folders[0]['id']

        return folder_id

    def create_folders_tree(self, path, root_folder_id):
        folder_id = root_folder_id
        for folder_name in path:
            folders = {folder['name']: folder['id'] for folder in self.list_folders(folder_name, folder_id)}

            if folder_name in folders:
                folder_id = folders[folder_name]
            else:
                folder_id = self.create_folder(folder_name, folder_id)

        return folder_id

    def create_folder(self, name, parent_folder_id=None):
        metadata = dict(name=name, mimeType='application/vnd.google-apps.folder')

        if parent_folder_id:
            metadata.update(dict(parents=[parent_folder_id]))

        self.log.debug('GoogleDriveApi: creating folder {} under {}'.format(name, parent_folder_id))
        return self.get_service().files().create(body=metadata, fields='id').execute().get('id')

    def list_folders(self, name, parent_folder_id):
        return self.get_service().files().list(
            q='\'{}\' in parents and name = \'{}\''.format(parent_folder_id, name)
        ).execute().get('files')
