import uuid
from falcon_rest.resources.resources import (
    ResourcesBaseClass, DEFAULT_SECURITY, errors, check_permission, responses, swagger
)
from falcon_rest.db.errors import IntegrityError, NoResultFound
from falcon_rest.conf import settings
from . import files_register


class UploadFile(ResourcesBaseClass):
    allow_methods = ['post']

    def __init__(
            self, scheme_class, entity='File', unique_field=None, additional_responses=None,
            path_parameters=None, security=None, body_params=None
    ):
        super(UploadFile, self).__init__()
        self.scheme_class = scheme_class
        self.entity = entity
        self.unique_field = unique_field
        self.additional_responses = (additional_responses or ())
        self.path_parameters = (path_parameters or ())
        self.security = (security if security is not None else DEFAULT_SECURITY)

        self.body_params = body_params

    def on_post(self, req, resp, **kwargs):
        model_class = self.scheme_class.Meta.model

        if self.check_request_data(req, resp, model_class.tmp_file_upload_scheme()):
            try:
                (
                    self.req_data['belongs_to_table'], self.req_data['belongs_to_field']
                ) = self.req_data['belongs_to'].split('.')

                self.req_data['path'] = self.place_file_and_return_path(req.files['file'])

                self.set_response(
                    resp, self.scheme_class.get_object_created_response(
                        scheme_class=model_class.tmp_file_scheme,
                        data=self.create_object(req, model_class.tmp_file_scheme(), **kwargs)
                    )
                )
            except PermissionError:
                self.set_response(resp, responses.OperationErrorResponse(
                    data=errors.CommonErrors.PermissionError
                ))
            except IntegrityError:
                self.set_response(resp, responses.AlreadyExistsResponse(
                    data=errors.CommonErrors.get_already_exists_error(self.entity, self.unique_field)
                ))

    def place_file_and_return_path(self, tmp_file):
        backend = files_register.get_backend(self.req_data['belongs_to_table'], self.req_data['belongs_to_field'])
        return backend.write(tmp_file)

    def get_spec(self):
        additional_responses = (
           responses.OperationErrorResponse(data=errors.CommonErrors.PermissionError, description='Can\'t create file'),
        ) + self.additional_responses

        ext_body_parameters = self.body_params or (
            {
                'name': 'belongs_to',
                'in': 'formData',
                'description': 'Entity and field which file should belong to',
                'required': True,
                'type': 'string',
                'enum': files_register.get_belongs_to_values(self.scheme_class)
            },
            {
                'name': 'public',
                'in': 'formData',
                'description': 'Whether file can be publicly available',
                'required': False,
                'type': 'boolean',
                'default': 'false'
            },
            {
                'name': 'file',
                'in': 'formData',
                'description': 'File to upload',
                'required': True,
                'type': 'file'
            }
        )

        return swagger.specify.get_spec(
            method='post', description='Creates new {}'.format(self.entity.lower()),
            consumes=['multipart/form-data'],
            path_parameters=self.path_parameters,
            responses=(
                self.scheme_class.get_object_created_response(entity=self.entity),
            ) + additional_responses + self.get_additional_responses(method='post'),
            security=self.get_security(method='post'),
            ext_body_parameters=ext_body_parameters
        )


class GetDownloadLink(ResourcesBaseClass):
    allow_methods = ['get']
    default_path_parameters = ()

    def __init__(
            self, scheme_class_req, scheme_class_resp, file_download_token_model_class, download_url='/file/download',
            belongs_to_pk_field_name='belongs_to_pk', entity='File', additional_responses=None,
            path_parameters=None, security=None
    ):
        super(GetDownloadLink, self).__init__()
        self.scheme_class_req = scheme_class_req
        self.scheme_class_resp = scheme_class_resp
        self.file_download_token_model_class = file_download_token_model_class
        self.download_url = download_url
        self.belongs_to_pk_field_name = belongs_to_pk_field_name
        self.entity = entity
        self.additional_responses = (additional_responses or ())
        self.path_parameters = (path_parameters or self.default_path_parameters)
        self.security = (security if security is not None else DEFAULT_SECURITY)

    def on_get(self, req, resp, **kwargs):
        if not check_permission(self, req, 'create'):
            self.set_response(
                resp, responses.ForbiddenErrorResponse(data=errors.AuthErrors.Forbidden)
            )
            return

        scheme = self.scheme_class_req()
        if self.check_path_parameters(req, resp, kwargs, scheme):
            file_obj = self.get_file_obj(req, resp, kwargs, scheme)
            if not file_obj:
                self.set_response(
                    resp, responses.ObjectNotFoundErrorResponse()
                )
                return

            self.set_response(
                resp,
                responses.SuccessResponseJustOk(
                    scheme=self.scheme_class_resp, data_key='download_link', data=self.get_download_link(file_obj.uuid)
                )
            )

    def get_download_token(self, file_uuid):
        token = '{}-{}-{}'.format(uuid.uuid4().hex, uuid.uuid4().hex, uuid.uuid4().hex)
        self.file_download_token_model_class(file_uuid=file_uuid, token=token).save()
        return token

    def get_download_link(self, file_uuid):
        token = self.get_download_token(file_uuid)
        return '{}{}{}{}/{}'.format(
            settings.API_SCHEME, settings.API_HOST, settings.API_BASE_PATH, self.download_url, token
        )

    # noinspection PyUnusedLocal,PyMethodMayBeStatic
    def get_file_obj(self, req, resp, data, scheme):
        file_obj = scheme.Meta.model.get(data['uuid'])
        if not file_obj:
            return None

        if file_obj.belongs_to_pk == data.get(self.belongs_to_pk_field_name):
            return file_obj
        else:
            return None

    def get_spec(self):
        additional_responses = (
            responses.SuccessResponseJustOk(scheme=self.scheme_class_resp, data_key='download_link'),
            responses.ObjectNotFoundErrorResponse(),
        ) + self.additional_responses

        return swagger.specify.get_spec(
            method='get', description='Gets file download link',
            path_parameters=self.path_parameters,
            responses=additional_responses + self.get_additional_responses(method='get'),
            security=self.get_security(method='get')
        )


class DownloadFile(ResourcesBaseClass):
    allow_methods = ['get']
    default_path_parameters = ()
    no_auth_needed = True

    def __init__(
            self, file_download_token_model_class, file_model_class, entity='File', additional_responses=None,
            path_parameters=None, security=None
    ):
        super(DownloadFile, self).__init__()
        self.file_download_token_model_class = file_download_token_model_class
        self.file_model_class = file_model_class
        self.entity = entity
        self.additional_responses = (additional_responses or ())
        self.path_parameters = (path_parameters or self.default_path_parameters)
        self.security = (security if security is not None else DEFAULT_SECURITY)

    def on_get(self, req, resp, **kwargs):
        if not check_permission(self, req, 'create'):
            self.set_response(
                resp, responses.ForbiddenErrorResponse(data=errors.AuthErrors.Forbidden)
            )
            return

        try:
            token_obj = self.file_download_token_model_class.filter(token=kwargs.get('download_token')).one()
        except NoResultFound:
            self.set_response(resp, responses.ObjectNotFoundErrorResponse())
            return

        file_obj = self.file_model_class.get(token_obj.file_uuid)

        if not file_obj:
            self.set_response(resp, responses.ObjectNotFoundErrorResponse())
            return

        backend = files_register.get_backend(file_obj.belongs_to_table, file_obj.belongs_to_field)

        resp.stream, resp.stream_len, filename, resp.content_type = backend.read(file_obj)
        resp.append_header('content-disposition', 'attachment; filename={}'.format(filename))

        if not settings.DONT_DELETE_DOWNLOAD_TOKEN:
            token_obj.delete()

    def get_spec(self):
        additional_responses = (responses.ObjectNotFoundErrorResponse(), ) + self.additional_responses

        return swagger.specify.get_spec(
            method='get', description='Downloads {}'.format(self.entity.lower()),
            produces=['application/json', 'application/octet-stream'],
            path_parameters=self.path_parameters,
            responses=additional_responses + self.get_additional_responses(method='get'),
            security=self.get_security(method='get')
        )


class ShowFile(ResourcesBaseClass):
    allow_methods = ['get']
    default_path_parameters = ()
    no_auth_needed = True

    def __init__(
        self, file_model_class, entity='File', additional_responses=None, path_parameters=None, security=None
    ):
        super(ShowFile, self).__init__()
        self.file_model_class = file_model_class
        self.entity = entity
        self.additional_responses = (additional_responses or ())
        self.path_parameters = (path_parameters or self.default_path_parameters)
        self.security = (security if security is not None else DEFAULT_SECURITY)

    def on_get(self, req, resp, **kwargs):
        if not check_permission(self, req, 'create'):
            self.set_response(
                resp, responses.ForbiddenErrorResponse(data=errors.AuthErrors.Forbidden)
            )
            return

        file_obj = self.file_model_class.get(kwargs.get('file_uuid'))

        if not file_obj or not file_obj.public:
            self.set_response(resp, responses.ObjectNotFoundErrorResponse())
            return

        backend = files_register.get_backend(file_obj.belongs_to_table, file_obj.belongs_to_field)

        resp.stream, resp.stream_len, filename, resp.content_type = backend.read(file_obj)
        resp.append_header('content-disposition', 'inline; filename={}'.format(filename))

    def get_spec(self):
        additional_responses = (responses.ObjectNotFoundErrorResponse(), ) + self.additional_responses

        return swagger.specify.get_spec(
            method='get', description='Shows {}'.format(self.entity.lower()),
            produces=['application/json', 'application/octet-stream'],
            path_parameters=self.path_parameters,
            responses=additional_responses + self.get_additional_responses(method='get'),
            security=self.get_security(method='get')
        )
