# noinspection PyProtectedMember
from falcon_rest.db import Column, fields, orm, errors, _BaseModelClass
from falcon_rest import schemes
from . backends import FileSystemBackEnd


class ModelUsingFiles(_BaseModelClass):
    file_fields_values_before_save = {}

    def before_update(self):
        self.file_fields_values_before_save = self.get_file_fields_values()

    def before_save(self):
        # noinspection PyAttributeOutsideInit
        self.autocommit = False
        file_fields_values = self.get_file_fields_values()
        tmp_files_to_move = {}
        # files_in
        for k, v in file_fields_values.items():
            if self.file_fields_values_before_save.get(k) != v:
                if self.get_field(k).file_model_class.get(v) is None:
                    tmp_files_to_move[k] = v

        self.move_tmp_files(tmp_files_to_move)
        # self.delete()

    def after_save(self):
        # noinspection PyUnresolvedReferences
        self.commit()

    def get_file_fields_values(self):
        ret = {}
        for column in self.__class__.__table__.columns:
            if isinstance(column, FileModelField):
                ret[column.name] = getattr(self, column.name)

        return ret

    def move_tmp_files(self, tmp_files_to_move):
        for field_name, file_uuid in tmp_files_to_move.items():
            file_model_class = self.get_field(field_name).file_model_class
            tmp_file = file_model_class.tmp_model_class.get(file_uuid)

            if not tmp_file:
                raise schemes.fields.ValidationError(
                    {field_name: 'File with UUID {input} does not exist'.format(input=file_uuid)}
                )

            file_model_class(
                uuid=tmp_file.uuid,
                path=tmp_file.path,
                belongs_to_table=tmp_file.belongs_to_table,
                belongs_to_field=tmp_file.belongs_to_field,
                belongs_to_pk=self.get_object_primary_key_value(),
                uploaded_on=tmp_file.uploaded_on,
                public=tmp_file.public
            ).save()

            tmp_file.delete()

    def delete_files(self, files_to_delete):
        pass


def relationship(
        file_model_class, model_class_tablename, file_column_name, model_class_name=None, relation_column_name=None
):
    if not model_class_name:
        model_class_name = '{}Model'.format(''.join([item.title() for item in model_class_tablename.split('_')]))

    file_model_class_name = file_model_class.__name__

    if not relation_column_name:
        relation_column_name = file_column_name.replace('_file_uuid', '')

    join_condition_str = ', '.join([
        'and_({model_class_name}.{file_column_name} == {file_model_class_name}.uuid',
        '{file_model_class_name}.belongs_to_table == "{model_class_tablename}"',
        '{file_model_class_name}.belongs_to_field == "{relation_column_name}")'
    ])
    join_condition_params = dict(
        model_class_name=model_class_name, file_column_name=file_column_name,
        model_class_tablename=model_class_tablename, file_model_class_name=file_model_class_name,
        relation_column_name=relation_column_name
    )

    return orm.relationship(file_model_class, primaryjoin=join_condition_str.format(**join_condition_params))


# noinspection PyAbstractClass
class FileModelField(Column):
    def __init__(self, file_model_class, backend=None, *args, **kwargs):
        if not args:
            args = (fields.String(36), )

        args += fields.ForeignKey('file.uuid'),

        super(FileModelField, self).__init__(*args, **kwargs)
        self.backend = (backend or FileSystemBackEnd())
        self.file_model_class = file_model_class


class FileSchemeField(schemes.fields.String):

    class Validator(schemes.fields.Validator):
        message_required = 'This field is required'
        message_invalid_uuid = 'UUID {input} is incorrect'
        message_file_not_exists = 'File with UUID {input} does not exist'

        def __init__(self, file_model_class, belongs_to_table, belongs_to_field, required=False):
            self.file_model_class = file_model_class
            self.belongs_to_table = belongs_to_table
            self.belongs_to_field = belongs_to_field
            self.required = required

        def _repr_args(self):
            return 'file_model_class={0!r}'.format(self.file_model_class)

        def __call__(self, value):
            if not value and not self.required:
                return value
            if not value:
                raise schemes.fields.ValidationError(self.message_required.format(input=value))
            if len(value) != 36:
                raise schemes.fields.ValidationError(self.message_invalid_uuid.format(input=value))

            try:
                self.file_model_class.filter(
                    belongs_to_table=self.belongs_to_table, belongs_to_field=self.belongs_to_field
                ).first()
            except errors.NoResultFound:
                try:
                    self.file_model_class.tmp_model_class.filter(
                        belongs_to_table=self.belongs_to_table, belongs_to_field=self.belongs_to_field
                    ).one()
                except errors.NoResultFound:
                    raise schemes.fields.ValidationError(self.message_file_not_exists.format(input=value))

            return value

    def __init__(self, file_model_class, belongs_to_table, belongs_to_field, *args, **kwargs):
        self.file_uuid_validator = self.Validator(file_model_class, belongs_to_table, belongs_to_field)
        super(FileSchemeField, self).__init__(*args, validate=self.file_uuid_validator, **kwargs)
        self.file_uuid_validator.required = self.required
