from . rbac_role import RbacRoleAnyRole, RbacRoleNoRole
import collections


class RbacAcl(object):
    _resources = []
    _roles = []
    _rules_list = {}
    default_perm = None

    def __init__(self, default_perm, no_role=RbacRoleNoRole, security_conf=None, additional_responses_conf=None):
        self.default_perm = default_perm
        self.no_role = no_role
        self.security_conf = security_conf
        self.additional_responses_conf = additional_responses_conf

    def is_allowed(self, resource, method, user=None, role=None, obj=None):
        if resource.rbac_hash() not in self._rules_list:
            return self.get_default_perm()

        if not role:
            role = user.get_role() if user else self.no_role

        if role.rbac_hash() not in self._roles:
            return self.get_default_perm()

        if role.rbac_hash() not in self._rules_list[resource.rbac_hash()]:
            """
            if role.rbac_hash() != 'any' and no explicit acl set for given role, calling
            is_allowed once more with role = RbacRoleAnyRole'
            """
            return self.is_allowed(
                resource, method, role=RbacRoleAnyRole
            ) if role is not RbacRoleAnyRole else self.default_perm

        if method != 'any' and method not in self._rules_list[resource.rbac_hash()][role.rbac_hash()]:
            """
            if method != 'any' and no explicit acl set for given method, calling is_allowed with method set to 'any'
            """
            return self.is_allowed(resource, 'any', role=role)

        return self._rules_list[resource.rbac_hash()][role.rbac_hash()].get(
            method, self.get_default_perm()
        )

    def add_rule(self, resource, perm, role=None, method=None, responses=None):
        self.init_resource(resource, responses)

        if resource.rbac_hash() not in self._resources:
            self._resources.append(resource.rbac_hash())
            self._rules_list[resource.rbac_hash()] = {}

        if role.rbac_hash() not in self._roles:
            self._roles.append(role.rbac_hash())

        if role.rbac_hash() not in self._rules_list[resource.rbac_hash()]:
            self._rules_list[resource.rbac_hash()][role.rbac_hash()] = {}

        self._rules_list[resource.rbac_hash()][role.rbac_hash()][method.lower() if method else 'any'] = perm

    def init_resource(self, resource, responses=None):
        if not hasattr(resource, 'rbac_hash'):
            resource.rbac_hash = lambda *args: resource.__name__

        if not hasattr(resource, 'acl'):
            resource.acl = self

        if not hasattr(resource, '__responses'):
            resource.__responses = responses

        if self.security_conf:
            if hasattr(resource, 'use_own_get_security') and getattr(resource, 'use_own_get_security') is True:
                return
            resource.use_own_get_security = False
            resource.get_security = lambda obj, **kwargs: self.get_resource_security(resource, kwargs['method'])

        if self.additional_responses_conf and (
            not hasattr(resource, 'get_additional_responses') or not resource.get_additional_responses(resource)
        ):
            resource.get_additional_responses = lambda obj, **kwargs: self.get_additional_responses(
                resource, kwargs.get('method')
            )

    def get_resource_security(self, resource, method):
        resource_security = []
        for role_class, swagger_security in self.security_conf:
            if self.is_allowed(resource, method, role=role_class).check():
                resource_security.append(swagger_security)

        return resource_security

    def get_additional_responses(self, resource, method=None):
        responses = ()

        for conf in self.additional_responses_conf:
            if 'allow' in conf:
                if not self.is_allowed(resource, method, role=conf['allow']).check():
                    continue

            if 'forbidden' in conf:
                if self.is_allowed(resource, method, role=conf['forbidden']).check():
                    continue

            responses = responses + conf['responses']

        if resource.__responses:
            responses = responses + resource.__responses

        return responses

    def add_rule_multiple(self, resources, perm, roles=None, methods=None, responses=()):
        for resource in resources:
            if roles and isinstance(roles, collections.Iterable):
                for role in roles:
                    if methods and isinstance(methods, collections.Iterable):
                        for method in methods:
                            self.add_rule(resource=resource, role=role, method=method, perm=perm, responses=responses)
                    else:
                        self.add_rule(resource=resource, role=role, perm=perm, responses=responses)
            elif methods and isinstance(methods, list):
                for method in methods:
                    self.add_rule(resource=resource, method=method, perm=perm, responses=responses)
            else:
                self.add_rule(resource=resource, perm=perm, responses=responses)

    def get_default_perm(self):
        if self.default_perm is not None:
            return self.default_perm
        else:
            raise NotImplementedError('Default permission did not set')
