from . import get_db
from . errors import IntegrityError, FkConstraintViolation
from . orm import get_foreign_keys

INTEGRITY_ERROR_VIOLATES_FOREIGN_KEY_CONSTRAINT = 'violates foreign key constraint'
INTEGRITY_ERROR_UNIQUE_ERROR_CODE = '23505'


class BaseModelClass(object):
    db_label = 'default'

    autocommit = True

    @classmethod
    def filter(cls, *args, **kwargs):
        fields = cls.get_fields()
        remove_keys = []
        if not args and kwargs:
            for k, v in kwargs.items():
                if k not in fields:  # pragma: no cover
                    continue
                args = args + (getattr(cls, k) == v, )
                remove_keys.append(k)

            for k in remove_keys:
                del kwargs[k]
        result = cls.query().filter(*args, **kwargs)
        return result

    @classmethod
    def query(cls):
        return get_db(cls.db_label).session.query(cls)

    @classmethod
    def get(cls, primary_key):
        return cls.query().get(primary_key)

    @classmethod
    def get_model_class_primary_key(cls):
        return get_db(cls.db_label).get_model_class_primary_key(cls)

    def before_save(self):
        pass

    def after_save(self):
        pass

    def save(self, **kwargs):
        self.before_save()

        try:
            result = get_db(self.db_label).save_object(self)
        except IntegrityError as e:
            #raise e
            if e.orig.pgcode == INTEGRITY_ERROR_UNIQUE_ERROR_CODE:
                if hasattr(self, 'after_save_fallback'):
                    self.after_save_fallback()
                raise e
            raise FkConstraintViolation(*self.scrutinize_integrity_error(e))

        if kwargs.get('commit', False):
            get_db(self.db_label).commit()

        self.after_save()

        return result

    def commit(self):
        get_db(self.db_label).commit()

    def delete(self, **kwargs):
        res = get_db(self.db_label).delete_object(self)
        if self.autocommit:
            self.commit()
        return res

    @classmethod
    def get_object_by_primary_key(cls, **kwargs):
        return get_db(cls.db_label).get_object_by_primary_key(
            cls, **{k: v for k, v in kwargs.items() if k in cls.get_fields()}
        )

    @classmethod
    def get_object_by_id(cls, id_value, **kwargs):
        return get_db(cls.db_label).get_object_by_id(cls, id_value)

    @classmethod
    def get_objects_list(cls, query=None, filtering=None, paging=None,
                         ordering=None):
        qs = query or cls.query()

        if filtering:
            for k, v in filtering.items():
                if '*' in v:  # pragma: no cover
                    qs = qs.filter(getattr(cls, k).like(v.replace('*', '%')))
                else:
                    qs = qs.filter(getattr(cls, k) == v)

        if ordering:  # pragma: no cover
            order_by = getattr(cls, ordering['by'])
            if ordering['dir'] == 'desc':
                order_by = order_by.desc()

            qs = qs.order_by(order_by)

        try:
            cnt = qs.count()
        except AssertionError:  # TODO: Fix this somehow
            cnt = len(qs.all())

        if paging:
            if 'from' in paging:
                qs = qs.offset(paging['from'])
            if 'till' in paging:
                qs = qs.limit(paging['till'] - paging.get('from', 0))

        return qs.all(), cnt

    @classmethod
    def get_name(cls):
        return cls.__name__.replace('Model', '')

    @classmethod
    def get_fields(cls):
        # noinspection PyUnresolvedReferences
        return [str(field).rsplit('.')[1] for field in cls.__table__.columns]

    @classmethod
    def get_field(cls, field_name):
        # noinspection PyUnresolvedReferences
        for field in cls.__table__.columns:
            if field.name == field_name:
                return field

    def get_object_primary_key_value(self):
        return get_db(self.db_label).get_object_primary_key_value(self)

    def scrutinize_integrity_error(self, error):
        error_msg = str(error.orig)
        params = error.params
        for fk in get_foreign_keys(self):
            if INTEGRITY_ERROR_VIOLATES_FOREIGN_KEY_CONSTRAINT in error_msg:
                if fk.parent.name in error_msg and fk.column.name in error_msg:
                    return fk.column.table.name, fk.column.name, fk.parent.name, params[fk.parent.name]
        return error_msg,'','',''###novvv
        ###raise Exception(error_msg ) #[msg for msg in error_msg.split('\n') if 'DETAIL:' in msg] or error_msg )
        # print('foreign_keys: {}'.format(foreign_keys))

    def __repr__(self):
        # return 'obj'
        parts = []
        for column_name in self.get_fields():
            parts.append('{} = {}'.format(column_name, getattr(self, column_name)))

        return '<{}: {}>'.format(self.__class__.__name__, ', '.join(parts))

