import re
import sqlalchemy
from sqlalchemy.inspection import inspect
from .. logger import log
from sqlalchemy.orm import scoped_session, sessionmaker


class DB(object):
    engine = None
    session = None
    label = 'default'

    def __init__(self, declarative_base, label='default'):
        self.engine = None
        self.session = None
        self.declarative_base = declarative_base
        self.label = label

    def save_object(self, obj):
        log.debug('DB [{}] saving object: {}'.format(self.label, obj))
        try:
            self.session.add(obj)
            self.session.commit()
            self.session.flush()
        except Exception as e:
            self.session.rollback()
            log.error('DB [{}]: error occurred: {}'.format(self.label, e))
            raise e
        return self.get_object_primary_key_value(obj)

    def delete_object(self, obj):
        log.debug('DB [{}] deleting object: {}'.format(self.label, obj))
        # noinspection PyBroadException
        try:
            self.session.delete(obj)
            return True
        except Exception as e:  # pragma: no cover
            self.session.rollback()
            log.error('DB [{}]: error occurred: {}'.format(self.label, e))
            raise e

    def get_object_by_primary_key(self, model_class, **kwargs):
        qs = self.session.query(model_class)
        try:
            for k, v in kwargs.items():
                qs = qs.filter(getattr(model_class, k) == v)
            return qs.one()
        except Exception as e:  # pragma: no cover
            self.session.rollback()
            log.error('DB: error occurred: {}'.format(e))
            raise e

    def get_object_by_id(self, model_class, id_value):
        return self.session.query(model_class).get(id_value)

    @staticmethod
    def get_object_primary_key_value(obj):
        pk = inspect(obj).identity
        if not pk:
            return None

        return pk[0] if len(pk) == 1 else pk

    @staticmethod
    def get_model_class_primary_key(model_class):
        primary_key = inspect(model_class).primary_key
        if len(primary_key) == 1:
            return primary_key[0].key
        else:  # pragma: no cover
            return tuple(pk.key for pk in primary_key)

    def create_engine(self, db_conn_string):
        log.info('DB [{}]: creating engine with {}'.format(self.label, re.sub(r':[^@]+@', ':********@', db_conn_string)))
        self.engine = sqlalchemy.create_engine(db_conn_string)

    def create_session(self):
        log.info('DB [{}]: creating session'.format(self.label))
        self.session = scoped_session(sessionmaker(bind=self.engine))

    def create_tables(self):
        log.info('DB [{}]: creating tables: {}'.format(self.label, list(self.declarative_base.metadata.tables.keys())))
        self.declarative_base.metadata.create_all(self.engine)
        log.info('DB [{}]: tables created'.format(self.label))

    def drop_tables(self):
        log.info('DB [{}]: dropping tables'.format(self.label))
        self.commit()
        self.declarative_base.metadata.drop_all(self.engine)

    def commit(self):
        self.session.commit()
