from urllib.parse import parse_qsl

import falcon
import simplejson as json
from falcon_rest.db import errors
from falcon_rest import conf, responses

def OperationalError(e):
    msg=str(e).split('\n')
    if len(msg) > 3:
      msg=msg[0:2]
    return responses.OperationErrorResponse(data=responses.errors.Error(40,e.__class__.__name__,msg ) )



class BaseResource(object):
    allow_methods = []
    has_update_by = False

    def __init__(self):
        self.req = None
        self.req_data = None
        self.params = None

    def init_req(self, req):
        self.req = req
        self.req_data = req.data

    def check_request_data(self, req, resp, scheme_obj):
        self.init_req(req)
        validation_errors = self.get_validation_errors(scheme_obj)

        if validation_errors:
            self.set_response(resp, responses.ValidationErrorResponse(data=validation_errors))
            return False
        else:
            return True

    def check_path_parameters(self, req, resp, data, scheme_obj):
        self.init_req(req)
        validation_errors = self.get_validation_errors(scheme_obj, data)

        if validation_errors:
            self.set_response(resp, responses.ValidationErrorResponse(data=validation_errors))
            return False
        else:
            return True

    def get_object(self, resp, model_class, **kwargs):
        primary_key = model_class.get_model_class_primary_key()
        if primary_key not in kwargs:#try to get synonyms
            from sqlalchemy import inspect
            for sn,sv in inspect(model_class).synonyms.items():
                if sv.name == primary_key and sn in kwargs:
                        kwargs[primary_key]=kwargs[sn]
                
        if isinstance(primary_key, str):  # pragma: no cover
            if primary_key not in kwargs:
                self.set_response(
                    resp,
                    responses.ValidationErrorResponse(data={primary_key: '{} not passed in path'.format(primary_key)})
                )
                return None
        else:  # pragma: no cover
            for pk in primary_key:
                if pk not in kwargs:
                    self.set_response(
                        resp,
                        responses.ValidationErrorResponse(
                            data={primary_key: '{} not passed in path'.format(pk)})
                    )
                    return None

        try:
            return model_class.get_object_by_primary_key(**kwargs)
        except errors.NoResultFound:
            self.set_response(resp, responses.ObjectNotFoundErrorResponse())
            return None
        except Exception as e:
                self.set_response( resp,OperationalError(e) )

    def get_object_data(self, resp, model_class, scheme_class, **kwargs):
        obj = self.get_object(resp, model_class, **kwargs)
        if not obj:
            return None
        else:
            return scheme_class().dump(obj).data

    def before_update(self, obj, req):
        if self.has_update_by:
            obj.update_by = self.get_user(req).name

        if hasattr(obj, 'before_update'):
            obj.before_update()

        return obj

    def after_update(self, result, obj, data):
        pass

    def update_object(self, req, resp, model_class, scheme_class, **kwargs):
        self.init_req(req)
        obj = self.get_object(resp, model_class, **kwargs)
        if not obj:
            return None

        scheme = scheme_class().load(self.req_data, instance=self.before_update(obj, req), partial=True)
        if scheme.errors:
            self.set_response(resp, responses.ValidationErrorResponse(data=scheme.errors))
            return False

        result = scheme.data.save(save_history=True, user=self.get_user(req))

        self.after_update(result, obj, scheme.data)

        return result

    def delete_object(self, req, resp, model_class, **kwargs):
        obj = self.get_object(resp, model_class, **kwargs)
        if not obj:
            return None
        return obj.delete(save_history=True, user=self.get_user(req))

    # noinspection PyMethodMayBeStatic,PyUnusedLocal
    def get_filtering_for_list(self, parsed_qs, **kwargs):
        filtering = {}
        for k, v in parsed_qs.items():
            if k in ['per_page', 'page', 'order_by', 'order_dir']:
                continue
            filtering[k] = v

        return filtering

    #@staticmethod
    def modify_query_from_filtering_for_list(self,filtering, **kwargs):
        return filtering, None

    def get_objects_list(self, req, model_class, **kwargs):
        parsed_qs = dict(parse_qsl(req.query_string))
        try:
            per_page = int(parsed_qs.get('per_page'))
        except (ValueError, TypeError):
            per_page = conf.get_default_items_per_page()

        try:
            page = int(parsed_qs.get('page', 0))
        except ValueError:
            page = 0

        if 'order_by' in parsed_qs:
            ordering = {
                'by': parsed_qs['order_by'],
                'dir': parsed_qs.get('order_dir', 'asc')
            }
        else:
            ordering = {}

        filtering = self.get_filtering_for_list(parsed_qs, **kwargs)

        filtering, query = \
            self.modify_query_from_filtering_for_list(filtering, **kwargs)

        return model_class.get_objects_list(
            query=query,
            filtering=filtering,
            ordering=ordering,
            paging={'from': page * per_page, 'till': (page + 1) * per_page}
        ) + (page, per_page)

    @staticmethod
    def set_response(resp, response_obj):
        resp.status = response_obj.falcon_status_code
        resp.body = json.dumps(response_obj.get_response_body())

    # def set_forbidden_response(self, resp):
    #     self.set_response(resp, ForbiddenErrorResponse(data=AuthErrors.AuthTokenNotProvided))

    def get_validation_errors(self, scheme, data=None):
        return scheme.validate(data or self.req_data)

    def get_loaded_data(self, scheme):
        return scheme.load(self.req_data).data

    # noinspection PyMethodMayBeStatic
    def get_user(self, req=None):
        return (req or self.req).context.get('user')

    # noinspection PyUnusedLocal,PyMethodMayBeStatic
    def before_create(self, obj, **kwargs):
        return obj

    def create_object(self, req, scheme, **kwargs):
        return self.before_create(
            self.get_loaded_data(scheme), **kwargs
        ).save(save_history=True, user=self.get_user(req))

    @staticmethod
    def method_not_allowed_response(resp, method):  # pragma: no cover
        resp.body = json.dumps({'error': 'Method {} not allowed'.format(method)}) + '\n'
        resp.status = falcon.HTTP_405

    # noinspection PyUnusedLocal
    def on_get(self, req, resp, **kwargs):  # pragma: no cover
        self.method_not_allowed_response(resp, 'GET')

    # noinspection PyUnusedLocal
    def on_post(self, req, resp, **kwargs):  # pragma: no cover
        self.method_not_allowed_response(resp, 'POST')

    # noinspection PyUnusedLocal
    def on_put(self, req, resp, **kwargs):  # pragma: no cover
        self.method_not_allowed_response(resp, 'PUT')

    # noinspection PyUnusedLocal
    def on_patch(self, req, resp, **kwargs):  # pragma: no cover
        self.method_not_allowed_response(resp, 'PATCH')

    # noinspection PyUnusedLocal
    def on_delete(self, req, resp, **kwargs):  # pragma: no cover
        self.method_not_allowed_response(resp, 'DELETE')

    # noinspection PyUnusedLocal
    def on_options(self, req, resp, *args, **kwargs):  # pragma: no cover
        resp.append_header('Allow', ', '.join(self.allow_methods))
