import re

from falcon_rest.db.errors import IntegrityError, FkConstraintViolation, DataError
from falcon_rest import schemes, swagger, responses
from falcon_rest.responses import errors
from falcon_rest.helpers import check_permission
from . base_resource import BaseResource,OperationalError
from falcon_rest.logger import log


ATTRIBUTE_ERROR_RE = re.compile(r'\'([^\']*)\'$')
DEFAULT_SECURITY = ['auth_token']


class RestrictRule(object):
    def __init__(self, user_model, user_object_field, data_field=None):
        self.user_model = user_model
        self.user_object_field = user_object_field
        self.data_field = data_field if data_field else user_object_field

    def apply(self, user, req_data, kwargs):
        if isinstance(user, self.user_model):
            req_data[self.data_field] = getattr(user, self.user_object_field)
            kwargs[self.data_field] = getattr(user, self.user_object_field)


class RestrictedModel(object):

    restrict = ()

    def can(self, req, permission):
        for rule in self.restrict:
            user = req.context['user']
            if isinstance(user, rule.user_model):
                if permission in ('create', 'delete'):
                    if getattr(self, rule.data_field) != getattr(user, rule.user_object_field):
                        return False

        return True


class ResourcesBaseClass(BaseResource):
    path_parameters = ()
    scheme_class = None
    id_field = None
    unique_field = None
    security = DEFAULT_SECURITY

    restrict = ()

    def get_security(self, **kwargs):
        return self.security

    # noinspection PyMethodMayBeStatic,PyUnusedLocal
    def get_additional_responses(self, **kwargs):
        return ()

    @classmethod
    def get_path_parameters(cls):
        if cls.id_field:
            if isinstance(cls.id_field, str):
                return cls.path_parameters + (
                    {'name': cls.id_field, 'description': '{} id to get info about'.format(cls.entity)},
                )
            elif isinstance(cls.id_field, tuple):  # pragma: no cover
                return cls.path_parameters + tuple(
                    {'name': field, 'description': '{} {} to get info about'.format(cls.entity, field)}
                    for field in cls.id_field
                )

        else:  # pragma: no cover
            return cls.path_parameters

    def apply_restrict_rules(self, req, kwargs):
        for rule in self.restrict:
            rule.apply(self.get_user(req), req.data or self.req_data, kwargs)


class Create(ResourcesBaseClass):
    allow_methods = ['post']

    scheme_class = None
    entity = None
    additional_responses = ()

    @classmethod
    def create(
        cls, scheme_class, entity, unique_field=None, additional_responses=None, path_parameters=None, security=None,
        restrict=()
    ):
        return type(
            '{}.Create'.format(scheme_class.__module__), (cls, ),
            dict(
                scheme_class=scheme_class, entity=entity, unique_field=unique_field,
                additional_responses=(additional_responses or ()),
                path_parameters=(path_parameters or ()),
                security=(security if security is not None else DEFAULT_SECURITY),
                restrict=restrict
            ),
        )

    def on_post(self, req, resp, **kwargs):
        self.apply_restrict_rules(req, kwargs)
        scheme = self.scheme_class()
        if not check_permission(self, req, 'create'):
            self.set_response(
                resp, responses.ForbiddenErrorResponse(data=errors.AuthErrors.Forbidden)
            )
            return

        if self.check_request_data(req, resp, scheme):
            try:
                self.set_response(
                    resp, self.scheme_class.get_object_created_response(data=self.create_object(req, scheme, **kwargs))
                )
            except IntegrityError:
                self.set_response(resp, responses.AlreadyExistsResponse(
                    data=errors.CommonErrors.get_already_exists_error(self.entity, self.unique_field))
                )
            except FkConstraintViolation as e:
                self.set_response(resp, responses.FkConstraintViolationResponse(
                    data=errors.CommonErrors.get_fk_violation_error(
                        e.table_name, e.fk_column_name, e.column_name, e.value)
                )
                )
            except schemes.validate.ValidationError as e:
                self.set_response(resp, responses.ValidationErrorResponse(data=e.messages))

            except DataError as e:
                log.error(e)
                #self.set_response(resp, responses.OperationErrorResponse(data=errors.CommonErrors.DataError))
                self.set_response( resp,OperationalError(e) )

    def get_spec(self):
        if self.unique_field:
            additional_responses = self.additional_responses + (
                responses.AlreadyExistsResponse(entity=[self.entity, self.unique_field]),
            )
        else:  # pragma: no cover
            additional_responses = self.additional_responses

        return swagger.specify.get_spec(
            method='post', description='Creates new {}'.format(self.entity.lower()),
            path_parameters=self.path_parameters,
            body_parameters=('{} to create'.format(self.entity), self.scheme_class),
            responses=(
                self.scheme_class.get_object_created_response(entity=self.entity),
                responses.ValidationErrorResponse(),
                responses.FkConstraintViolationResponse()
            ) + additional_responses + self.get_additional_responses(method='post'),
            security=self.get_security(method='post')
        )


class Resource(ResourcesBaseClass):
    allow_methods = []

    model_class = None
    scheme_class_get = None
    entity = None
    has_update_by = False
    scheme_class_modify = None

    has_info_operation = True
    has_modify_operation = True
    has_delete_operation = True

    def get_common_on_get_responses(self):
        return (
            responses.SuccessResponseObjectInfo(payload_scheme=self.scheme_class_get),
            responses.ObjectNotFoundErrorResponse()
        )

    def __init__(self, has_info_operation=True, has_modify_operation=True, has_delete_operation=True):
        super(Resource, self).__init__()
        self.allow_methods = []

        if has_info_operation:
            self.allow_methods.append('get')
        else:
            self.has_info_operation = False

        if has_modify_operation:
            self.allow_methods.append('patch')
        else:
            self.has_modify_operation = False

        if has_delete_operation:
            self.allow_methods.append('delete')
        else:
            self.has_delete_operation = False

    def get_security(self, **kwargs):
        if not isinstance(self.security, dict):
            return self.security

        method = kwargs['method']
        action = kwargs['action']

        if method in self.security:
            return self.security[method]

        if action in self.security:
            return self.security[action]

        raise NotImplementedError(
            'Can\'t determine swagger security from {} for action / method {} / {}'.format(
                self.security, method, action
            )
        )

    @classmethod
    def create(
            cls, model_class, scheme_class, scheme_class_get, entity, id_field, has_update_by=False, unique_field=None,
            scheme_class_modify=None, security=None, restrict=()
    ):
        return type(
            '{}.Resource'.format(model_class.__module__), (Resource, ),
            dict(
                model_class=model_class, scheme_class=scheme_class, scheme_class_get=scheme_class_get,
                scheme_class_modify=scheme_class_modify,
                entity=entity, id_field=id_field,
                has_update_by=has_update_by, unique_field=unique_field,
                security=(security if security is not None else DEFAULT_SECURITY),
                restrict=restrict
            )
        )

    def on_get(self, req, resp, **kwargs):
        if not self.has_info_operation:  # pragma: no cover
            return self.method_not_allowed_response(resp, 'GET')

        if not check_permission(self, req, 'show', self.get_object(resp, self.model_class, **kwargs)):
            self.set_response(
                resp, responses.ForbiddenErrorResponse(data=errors.AuthErrors.Forbidden)
            )
            return

        data = self.get_object_data(resp, self.model_class, self.scheme_class_get, **kwargs)
        if data:
            self.set_response(resp, responses.SuccessResponseObjectInfo(data=data))

    def get_spec_info(self):
        return swagger.specify.get_spec(
            method='get', description='Gets {}'.format(self.entity.lower()),
            path_parameters=self.get_path_parameters(),
            responses=self.get_common_on_get_responses() + self.get_additional_responses(method='get'),
            security=self.get_security(method='get', action='info')
        )

    def on_patch(self, req, resp, **kwargs):
        if not self.has_modify_operation:  # pragma: no cover
            return self.method_not_allowed_response(resp, 'PATCH')

        if not check_permission(self, req, 'modify', self.get_object(resp, self.model_class, **kwargs)):
            self.set_response(
                resp, responses.ForbiddenErrorResponse(data=errors.AuthErrors.Forbidden)
            )
            return

        try:
            if self.update_object(
                req, resp, self.model_class,
                self.scheme_class_modify if self.scheme_class_modify else self.scheme_class, **kwargs
            ):
                self.set_response(resp, responses.SuccessResponseJustOk())
        except IntegrityError:
            self.set_response(resp, responses.AlreadyExistsResponse(
                data=errors.CommonErrors.get_already_exists_error(self.entity, self.unique_field))
            )
        except FkConstraintViolation as e:
            self.set_response(resp, responses.FkConstraintViolationResponse(
                data=errors.CommonErrors.get_fk_violation_error(e.table_name, e.column_name, e.value))
            )
        except Exception as e:
            log.error(e)
            #self.set_response(resp, responses.OperationErrorResponse(data=errors.CommonErrors.DataError))
            self.set_response(resp, OperationalError(e) )

    def get_spec_modify(self):
        if self.scheme_class or self.scheme_class_modify:
            body_parameters = (
                '{} to modify'.format(self.entity),
                self.scheme_class_modify if self.scheme_class_modify else self.scheme_class

            )
        else:
            body_parameters = ()

        return swagger.specify.get_spec(
            method='patch', description='Modifies {}'.format(self.entity.lower()),
            path_parameters=self.get_path_parameters(),
            body_parameters=body_parameters,
            responses=(
                responses.SuccessResponseJustOk(),
                responses.ObjectNotFoundErrorResponse()
            ) + self.get_additional_responses(method='patch'),
            security=self.get_security(method='patch', action='modify')
        )

    def on_delete(self, req, resp, **kwargs):
        if not self.has_delete_operation:  # pragma: no cover
            return self.method_not_allowed_response(resp, 'DELETE')

        if not check_permission(self, req, 'delete', self.get_object(resp, self.model_class, **kwargs)):
            self.set_response(
                resp, responses.ForbiddenErrorResponse(data=errors.AuthErrors.Forbidden)
            )
            return

        if self.delete_object(req, resp, self.model_class, **kwargs):
            self.set_response(resp, responses.SuccessResponseJustOk())

    def get_spec_delete(self):
        return swagger.specify.get_spec(
            method='delete', description='Deletes {}'.format(self.entity.lower()),
            path_parameters=self.get_path_parameters(),
            responses=(
                responses.SuccessResponseJustOk(),
                responses.ObjectNotFoundErrorResponse()
            ) + self.get_additional_responses(method='delete'),
            security=self.get_security(method='delete', action='delete')
        )

    def get_spec(self):
        spec = {}
        if self.has_info_operation:
            spec.update(self.get_spec_info())

        if self.has_modify_operation:
            spec.update(self.get_spec_modify())

        if self.has_delete_operation:
            spec.update(self.get_spec_delete())

        return spec


class List(ResourcesBaseClass):
    allow_methods = ['get']
    entity_plural = None
    model_class = None

    @classmethod
    def create(cls, model_class, scheme_class, entity_plural, path_parameters=None, security=None, restrict=()):
        return type(
            '{}.List'.format(model_class.__module__), (cls, ), dict(
                scheme_class=scheme_class, model_class=model_class, entity_plural=entity_plural,
                path_parameters=(path_parameters or ()),
                security=(security if security is not None else DEFAULT_SECURITY),
                restrict=restrict
            ),
        )

    def on_get(self, req, resp, **kwargs):
        if not check_permission(self, req, 'list'):
            self.set_response(
                resp, responses.ForbiddenErrorResponse(data=errors.AuthErrors.Forbidden)
            )
            return
        try:
            objects_list, total, page, per_page = self.get_objects_list(req, self.model_class, **kwargs)
            self.set_response(
                resp, responses.SuccessResponse(
                    data={
                        'items': self.scheme_class().dump(objects_list, many=True).data,
                        'total': total, 'page': page, 'per_page': per_page
                    },
                    scheme=schemes.ObjectScheme
                )
            )
        except AttributeError as e:
            self.set_response(
                resp, responses.AttributeNotExitsErrorResponse(
                    data=ATTRIBUTE_ERROR_RE.search(str(e)).group(1)
                )
            )
        except Exception as e:
            log.error(e)
            #self.set_response(resp, responses.OperationErrorResponse(data=errors.CommonErrors.DataError))
            self.set_response(resp, OperationalError(e) )


    def get_spec(self):
        fields_names, search_fields_list = swagger.specify.get_fields_from_scheme(self.scheme_class)

        return swagger.specify.get_spec(
            method='get', description='Gets {}'.format(self.entity_plural.lower()), path_parameters=self.path_parameters,
            query_parameters=[
                {'name': 'page', 'type': 'integer'}, {'name': 'per_page', 'type': 'integer'},
                {'name': 'order_by', 'enum': fields_names}, {'name': 'order_dir', 'enum': ['asc', 'desc']}
            ] + search_fields_list,
            responses=(
                responses.SuccessResponseObjectsList(payload_scheme_items=self.scheme_class),
                responses.AttributeNotExitsErrorResponse()
            ) + self.get_additional_responses(method='get'),
            security=self.get_security(method='get')
        )


def create_entity_endpoints(
        model_class, scheme_class, scheme_class_get, entity_sp, id_field, unique_field=None, has_update_by=False,
        additional_responses=None, path_parameters=None, scheme_class_modify=None, security=None, restrict=()
):
    if hasattr(model_class, 'restrict'):
        restrict = model_class.restrict

    return (
        Create.create(
            scheme_class, entity_sp[0], unique_field, additional_responses, path_parameters=path_parameters,
            **(dict(security=security['create']) if isinstance(security, dict) and 'create' in security else {}),
            restrict=restrict
        ),
        Resource.create(
            model_class, scheme_class, scheme_class_get, entity_sp[0], id_field,
            has_update_by=has_update_by, unique_field=unique_field, scheme_class_modify=scheme_class_modify,
            **(dict(security=security['resource']) if isinstance(security, dict) and 'resource' in security else {}),
            restrict=restrict
        ),
        List.create(
            model_class, scheme_class_get, entity_sp[1], path_parameters=path_parameters,
            **(dict(security=security['list']) if isinstance(security, dict) and 'list' in security else {}),
            restrict=restrict
        )
    )


class PreConditionFailedException(Exception):
    def __init__(self, message):
        self.message = message


class CustomAction(ResourcesBaseClass):
    model_class = None
    method = None
    path_parameters = ()
    description = ''
    # field, value, error_message
    pre_conditions = ()
    # field, value, error_message
    pre_conditions_inverted = ()

    action = None
    allow_methods = []
    restrict = ()

    body_parameters = ()

    additional_responses = ()

    def __init__(self):
        super(CustomAction, self).__init__()
        self.allow_methods = [self.method]

        if not self.action:
            self.action = self.__class__.__name__.rsplit('.', 1)[0]

    def proceed(self, req, resp, **kwargs):

        if not check_permission(self, req, self.action, self.get_object(resp, self.model_class, **kwargs)):
            self.set_response(
                resp, responses.ForbiddenErrorResponse(data=errors.AuthErrors.Forbidden)
            )
            return

        try:
            obj = self.get_object(resp, self.model_class, **kwargs)
            if self.check_conditions(obj, self.pre_conditions, lambda a, b: a == b):
                if self.check_conditions(obj, self.pre_conditions_inverted, lambda a, b: a != b):
                    if self.apply(obj, req, resp, **kwargs):
                        self.set_response(resp, responses.SuccessResponseJustOk())

        except FkConstraintViolation as e:
            self.set_response(resp, responses.FkConstraintViolationResponse(
                data=errors.CommonErrors.get_fk_violation_error(e.table_name, e.column_name, e.value))
            )
        except PreConditionFailedException as e:
            self.set_response(resp, responses.OperationErrorResponse(
                data=errors.CommonErrors.get_pre_condition_failed_error(e.message))
            )

    @staticmethod
    def check_conditions(obj, conditions, cmp):
        for field, value, error_message in conditions:
            if not cmp(getattr(obj, field), value):
                raise PreConditionFailedException(error_message)

        return True

    def apply(self, obj, req, resp, **kwargs):
        raise NotImplementedError('This method must be implemented in any derived class.')

    def get_spec(self):
        return swagger.specify.get_spec(
            method=self.method, description=self.description,
            path_parameters=self.path_parameters,
            body_parameters=self.body_parameters,
            responses=(
                responses.SuccessResponseJustOk(),
                responses.ObjectNotFoundErrorResponse()
            ) + self.additional_responses,
            security=self.get_security(method=self.method)
        )


# noinspection PyAbstractClass
class CustomPatchAction(CustomAction):
    method = 'patch'

    def on_patch(self, req, resp, **kwargs):
        return self.proceed(req, resp, **kwargs)
