import falcon

from falcon_rest.schemes import schemes
from . errors import AuthErrors, CommonErrors


class BaseResponse(object):
    status_code = 200
    falcon_status_code = falcon.HTTP_200
    scheme = None
    description = ''
    data = None
    data_key = None
    tpl = None
    many = False
    payload_scheme = None
    payload_scheme_items = None

    def __init__(
            self, description=None, data=None, entity='', scheme=None, payload_scheme=None, payload_scheme_items=None,
            data_key=None
    ):
        if description is not None:
            self.description = description

        if data is not None:
            self.data = data

        if entity:
            self.description = self.tpl.format(*entity) if isinstance(entity, list) else self.tpl.format(entity)

        if scheme is not None:
            self.scheme = scheme

        if payload_scheme is not None:
            self.payload_scheme = payload_scheme

        if payload_scheme_items is not None:
            self.payload_scheme_items = payload_scheme_items

        if data_key:
            self.data_key = data_key

    def get_response_body(self):
        return self.scheme().dump({self.data_key: self.data}, many=self.many).data


class SuccessResponseJustOk(BaseResponse):
    description = 'Successfully executed'
    scheme = schemes.SuccessScheme


class SuccessResponse(BaseResponse):
    description = 'Successfully executed'
    data_key = 'payload'


class SuccessResponseObjectInfo(BaseResponse):
    description = 'Successfully executed'
    scheme = schemes.ObjectScheme
    data_key = 'payload'


class SuccessResponseObjectsList(SuccessResponse):
    # many = True
    scheme = schemes.ObjectsListScheme


class ObjectCreatedResponse(BaseResponse):
    scheme = schemes.ObjectCreated
    data_key = 'object_id'
    tpl = '{} successfully created'
#
#
# class ObjectCreatedUUIDResponse(BaseResponse):
#     scheme = schemes.ObjectCreatedUUID
#     data_key = 'object_uuid'
#     tpl = '{} successfully created'


class ValidationErrorResponse(BaseResponse):
    status_code = 400
    falcon_status_code = falcon.HTTP_400
    scheme = schemes.ValidationError
    description = 'Validation error'
    data_key = 'errors'


class AttributeNotExitsErrorResponse(BaseResponse):
    status_code = 406
    falcon_status_code = falcon.HTTP_406
    scheme = schemes.AttributeNotExists
    description = 'Attribute which was provided in query string for search by doesnt exist'
    data_key = 'attribute'


class ObjectNotFoundErrorResponse(BaseResponse):
    status_code = 404
    falcon_status_code = falcon.HTTP_404
    scheme = schemes.ObjectNotFoundError
    description = 'Object not found error'
    data_key = 'error'


class OperationErrorResponse(BaseResponse):
    status_code = 406
    falcon_status_code = falcon.HTTP_406
    scheme = schemes.OperationError
    data_key = 'error'


class AlreadyExistsResponse(OperationErrorResponse):
    tpl = '{} with such {} already exists'


class FkConstraintViolationResponse(OperationErrorResponse):
    tpl = '{} with {} = {} does not exist in the Database'


class UnAuthorizedErrorResponse(OperationErrorResponse):  # incorrect credentials
    status_code = 401
    falcon_status_code = falcon.HTTP_401
    description = 'Incorrect credentials'
    data = AuthErrors.IncorrectCredentials


class UnauthenticatedErrorResponse(OperationErrorResponse):  # not authorized
    description = 'You are not unauthenticated'
    status_code = 401
    falcon_status_code = falcon.HTTP_401
    scheme = schemes.UnauthenticatedError


class ForbiddenErrorResponse(OperationErrorResponse):  # not authorized
    description = 'You are not authorized to access this resource'
    status_code = 403
    falcon_status_code = falcon.HTTP_403
    scheme = schemes.ForbiddenError
    data = AuthErrors.Forbidden


class MalformedJsonResponse(OperationErrorResponse):
    description = 'Malformed JSON provided'
    status_code = 753
    falcon_status_code = falcon.HTTP_753
    scheme = schemes.MalformedJsonError
    data = CommonErrors.MalformedJson
