from marshmallow import fields
from falcon_rest.responses import UnauthenticatedErrorResponse
from falcon_rest.schemes import ObjectScheme, ObjectsListScheme
from collections import OrderedDict


CONSUMES = ['application/json']
PRODUCES = ['application/json']


def get_response(response):
    ret = OrderedDict([('description', response.description)])

    if response.scheme is ObjectScheme:
        ret['schema'] = {'properties': OrderedDict([
            ('success', {'type': 'boolean', 'default': True}),
            ('payload', {'$ref': '#/definitions/{}'.format(response.payload_scheme.__name__.replace('Scheme', ''))})
        ])}

    elif response.scheme is ObjectsListScheme:
        ret['schema'] = {'properties': OrderedDict([
            ('success', {'type': 'boolean', 'default': True}),
            ('payload', OrderedDict([
                ('type', 'object'),
                ('properties', OrderedDict([
                    ('total', {'type': 'integer', 'description': 'Total items found'}),
                    ('page', {'type': 'integer', 'description': 'Current page (starting from 0)'}),
                    ('per_page', {'type': 'integer', 'description': 'Items per page'}),
                    ('items', {
                        'type': 'array',
                        'items': {
                            '$ref': '#/definitions/{}'.format(
                                response.payload_scheme_items.__name__.replace('Scheme', ''))
                        }
                    })
                ]))
            ]))
            # {'$ref': '#/definitions/{}'.format(response.payload_scheme.__name__.replace('Scheme', ''))})
        ])}

    else:
        ret.update({'schema': {'$ref': '#/definitions/{}'.format(response.scheme.__name__.replace('Scheme', ''))}})

    return ret


def get_spec(
        method='get', description='', consumes=CONSUMES, produces=PRODUCES, body_parameters=(),
        path_parameters=(), query_parameters=(), responses=(), security=(), ext_body_parameters=()
):
    spec = {
        method: {
            'description': description,
            'consumes': consumes,
            'produces': produces,
            'responses': {
                r.status_code: get_response(r) for r in responses
            }
        }
    }

    if body_parameters:
        if 'parameters' not in spec[method]:
            spec[method]['parameters'] = []

        spec[method]['parameters'].append({
            'in': 'body',
            'name': 'body',
            'description': body_parameters[0],
            'schema': {
                '$ref': '#/definitions/{}'.format(body_parameters[1].__name__.replace('Scheme', ''))
            }
        })

    if ext_body_parameters:
        if 'parameters' not in spec[method]:
            spec[method]['parameters'] = []

        for param in ext_body_parameters:
            spec[method]['parameters'].append(param)

    if path_parameters:
        if 'parameters' not in spec[method]:
            spec[method]['parameters'] = []

        for param in path_parameters:
            spec[method]['parameters'].append({
                'in': 'path',
                'name': param['name'],
                'description': param.get('description', ''),
                'required': True,
                'type': param.get('type', 'integer' if '_id' in param['name'] or param['name'] == 'id' else 'string'),
                'format': param.get('format', 'int64')
            })

    if query_parameters:
        if 'parameters' not in spec[method]:
            spec[method]['parameters'] = []

        for param in query_parameters:
            p = {
                'in': 'query',
                'name': param['name'],
                'required': param.get('required', False),
                'type': param.get('type', 'string')
            }
            if 'description' in param:  # pragma: no cover
                p['description'] = param['description']

            if 'default' in param:  # pragma: no cover
                p['default'] = param['default']

            if 'enum' in param:
                p['enum'] = param['enum']

            spec[method]['parameters'].append(p)

    # tag = get_tag(fn)
    # if tag:
    #     spec['method']

    if security:
        spec[method]['security'] = []
        for sec in security:
            if isinstance(sec, dict):
                for name, scopes in sec:
                    spec[method]['security'].appens({name: scopes})
            else:  # string
                spec[method]['security'].append({sec: []})

        if 'auth_token' in security:
            for response_class in [UnauthenticatedErrorResponse]:
                response = response_class()
                spec[method]['responses'][response.status_code] = get_response(response)

    return spec


def get_field_type(field):  # pragma: no cover
    if isinstance(field, fields.Boolean):
        return 'boolean'
    elif isinstance(field, fields.Integer):
        return 'integer'
    elif isinstance(field, fields.Float):
        return 'float'
    elif isinstance(field, fields.String):
        return 'string'
    elif isinstance(field, fields.DateTime):
        return 'datetime'

    return None


def get_fields_from_scheme(scheme_class):  # pragma: no cover
    fields_list = []
    fields_names = []

    search_fields = scheme_class.Meta.search_fields if hasattr(scheme_class.Meta, 'search_fields') else []

    # noinspection PyProtectedMember
    for field_name, field in scheme_class._declared_fields.items():
        field_type = get_field_type(field)

        if not field_type or field_name in []:
            continue

        fields_names.append(field_name)

        if field_type == 'datetime' or field_name not in search_fields:
            continue

        f = {
            'name': field_name,
            'type': field_type
        }

        if isinstance(field.validate, fields.validate.OneOf):
            f['enum'] = field.validate.choices

        fields_list.append(f)

    return sorted(fields_names), sorted(fields_list, key=lambda item: item['name'])
