import glob
import json
import os
import sys
from types import ModuleType
from . apispec import APISpecWithSecurityDefinitions
import yaml

from falcon_rest import schemes

COMMON_SCHEMES = (
    schemes.SuccessScheme, schemes.UnauthenticatedError, schemes.ObjectNotFoundError, schemes.ObjectCreated,
    schemes.ValidationError, schemes.OperationError, schemes.AttributeNotExists
)

DEFAULT_OAUTH2_SD_DESCRIPTION = 'In fact it is password flow, but it\'s not currently supported by Swagger UI'


class EndPoints(object):
    def __init__(self, swagger):
        self.swagger = swagger

    def on_get(self, req, resp):
        if '.yaml' in req.path:
            resp.append_header('Content-Type', 'text/x-yaml')
            resp.body = self.swagger.get_yaml()
        else:
            resp.body = self.swagger.get_json()


class Swagger(object):
    def __init__(self, title, host, base_path, version):
        spec_config = {
            'title': title,
            'basePath': base_path,
            'version': version,
            'plugins': (
                'apispec.ext.marshmallow',
            )
        }

        if host:
            spec_config.update({'host': host})

        self.spec = APISpecWithSecurityDefinitions(**spec_config)

        self.endpoints = EndPoints(self)

    def get_end_points(self):
        return type('SwaggerEndPoints', (EndPoints, ), dict(swagger=self))

    def add_path(self, path, resource, tags, deprecated):
        self.spec.add_path(path, operations=resource.get_spec(), tags=tags, deprecated=deprecated)

    def add_paths(self, paths):
        for tag_name, tag in paths.items():
            if isinstance(tag, dict):
                self.spec.add_tag({'name': tag_name, 'description': tag['description']})
                routes_list = tag['routes']
            elif isinstance(tag, list):
                self.spec.add_tag({'name': tag_name})
                routes_list = tag

            for route in routes_list:
                self.add_path(route['path'], route['resource'], [tag_name], route.get('deprecated', False))

    def add_definition(self, scheme_to_add):
        self.spec.definition(scheme_to_add.__name__.replace('Scheme', ''), schema=scheme_to_add)

    def add_security_definition(self, name, definition):
        self.spec.security_definition(name, definition)

    def get_json(self):
        return json.dumps(self.spec.to_dict(), indent=4).replace('"default": "false"', '"default": false')

    def get_yaml(self):
        return yaml.dump(json.loads(json.dumps(self.spec.to_dict()))).replace('default: \'false\'', 'default: false')

    def add_definitions_from_module(self, mod, skip_pattern='Inner', take_pattern='Scheme'):
        if isinstance(mod, str):  # pragma: no cover
            __import__(mod)
            mod = sys.modules[mod]

        for name, cls in mod.__dict__.items():
            if isinstance(cls, ModuleType):
                for cls_name, subclass in cls.__dict__.items():
                    self.add_definitions_from_class(cls_name, subclass, skip_pattern, take_pattern)
            else:
                self.add_definitions_from_class(name, cls, skip_pattern, take_pattern)

    def add_definitions_from_class(self, name, cls, skip_pattern='Inner', take_pattern='Scheme'):
        if not isinstance(cls, type):
            return
        if (skip_pattern and skip_pattern in name) or (take_pattern and take_pattern not in name):
            return

        self.add_definition(cls)

    def add_definitions_from_submodules(self, mod_or_mod_name, skip_pattern='Inner', take_pattern='Scheme'):
        if isinstance(mod_or_mod_name, ModuleType):
            self.add_definitions_from_module(mod_or_mod_name, skip_pattern, take_pattern)
            return

        __import__(mod_or_mod_name)
        mod = sys.modules[mod_or_mod_name]
        self.add_definitions_from_module(mod, skip_pattern, take_pattern)
        if not mod.__file__.endswith('__init__.py'):
            return

        for sub_mod_file in glob.glob(os.path.join(os.path.dirname(mod.__file__), '*')):
            sub_mod_file_base_name = os.path.basename(sub_mod_file)
            if sub_mod_file_base_name.startswith('__'):
                continue

            sub_mod_name = '{}.{}'.format(mod_or_mod_name, os.path.splitext(sub_mod_file_base_name)[0])
            if os.path.isdir(sub_mod_name) or os.path.splitext(sub_mod_name)[1] == 'py' or True:
                try:
                    __import__(sub_mod_name)
                except ImportError:
                    print('WARNING: can\'t import {}'.format(sub_mod_name))
                    continue
            else:
                continue

            self.add_definitions_from_module(sys.modules[sub_mod_name], skip_pattern, take_pattern)

    def add_common_schemes(self):
        for scheme in COMMON_SCHEMES:
            self.add_definition(scheme)

    def add_oauth2_security_definition(self, name, url, description=DEFAULT_OAUTH2_SD_DESCRIPTION, scopes=None):
        definition = dict(type='oauth2', authorizationUrl=url, flow='implicit', description=description)
        if scopes:
            definition.update(dict(scopes=scopes))

        self.add_security_definition(name, definition)

    def add_api_key_security_definition(self, name, key_name='X-Auth-Token', key_in='header', description=''):
        self.add_security_definition(
            name, {'type': 'apiKey', 'name': key_name, 'in': key_in, 'description': description}
        )

    def register_path_helper(self, helper):
        self.spec.register_path_helper(helper)
